#version 430
/*
* Copyright 2019 Vienna University of Technology.
* Institute of Computer Graphics and Algorithms.
* This file is part of the ECG Lab Framework and must not be redistributed.
*/

in VertexData {
	vec3 position_world;
	vec3 normal_world;
	vec2 uv;
} vert;

out vec4 color;

uniform vec3 camera_world;

uniform vec3 materialCoefficients; // x = ambient, y = roughness, z = metalness 
uniform float specularAlpha;
uniform sampler2D diffuseTexture;
uniform sampler2D roughnessTexture;
uniform sampler2D metalnessTexture;
uniform sampler2D aoTexture;

uniform struct DirectionalLight {
	vec3 color;
	vec3 direction;
} dirL;

uniform struct PointLight {
	vec3 color;
	vec3 position;
	vec3 attenuation;
} pointL;

const float PI = 3.14159265359;
float roughness;
float metalness;
float ao;
vec3 R0;

//Beckmann distribution
//https://en.wikipedia.org/wiki/Specular_highlight#Beckmann_distribution
float beckmann(float nh) {
	float m = roughness;
	float alpha = acos(nh);
	return (exp((-pow(tan(alpha),2))/pow(m,2)))/max(0.000001, (PI*pow(m,2)*pow(cos(alpha),4)));
}

float dist_ggx(float nh){
	float a = roughness*roughness;
	float a2 = pow(a,2);
	float denom = pow(nh,2)*(a2-1.0)+1.0;
	denom = PI * pow(denom, 2);
	return a2 / max(0.0000001,denom);
}

//fresnel factor(schlicks approximation)
//https://en.wikipedia.org/wiki/Schlick%27s_approximation
vec3 fresnel(float hv) {
	return R0 + (1 - R0)*pow((1 - hv),5);
}


//geometric attenuation
//https://en.wikipedia.org/wiki/Specular_highlight#Cook.E2.80.93Torrance_model
float geom_att(float nh, float vn, float hv, float nl) {
	float first = (2*nh*vn)/(hv);
	float second = (2*nh*nl)/(hv);
	return min(1,min(first,second));
}

float geom_smith(float vn, float nl){
	float r = roughness + 1.0;
	float k = (r*r) / 8.0;
	float ggx1 = vn / (vn*(1.0-k)+k);
	float ggx2 = nl / (nl*(1.0-k)+k);
	return ggx1 * ggx2;
}

//cook-torrance-model
//https://en.wikipedia.org/wiki/Specular_highlight#Cook.E2.80.93Torrance_model
vec3 cookTorrance(vec3 n, vec3 l, vec3 v, vec3 h) {
	float nl = max(0.0f, dot(n, l));
	float nh = max(0.0f, dot(n, h));
	float vn = max(0.0f, dot(v, n));
	float hv = max(0.0f, dot(h, v));

	//float D = beckmann(nh);
	float D = dist_ggx(nh);
	vec3 F = fresnel(hv);
	//float G = geom_att(nh, vn, hv, nl);
	float G = geom_smith(vn, nl);

	return (D*F*G)/(max(0.000001, (4.0*vn*nl)));
}

vec3 dirLight(DirectionalLight light, vec3 n, vec3 v, vec3 texColor) {
	vec3 l = normalize(-light.direction);
	vec3 h = normalize(v+l);

	
	vec3 ks = fresnel(dot(h,v));
	vec3 kd = vec3(1.0)-ks;
	kd *= 1.0-metalness; 

	vec3 diff = kd * texColor/PI;
	vec3 spec = cookTorrance(n,l,v,h);

	return (diff+spec)*light.color*max(0,dot(n,l));
}

vec3 ptLight(PointLight light, vec3 n, vec3 v, vec3 texColor) {
	vec3 l = light.position-vert.position_world;
	float dist = length(l);  
	l = normalize(l);
	vec3 h = normalize(v + l);

	float att = 1.0 / max(1,(light.attenuation.x+light.attenuation.y*dist+light.attenuation.z*pow(dist,2)));

	vec3 ks = fresnel(dot(h,v));
	vec3 kd = vec3(1.0)-ks;
	kd *= 1.0-metalness; 

	vec3 diff = kd * texColor/PI;
	vec3 spec = cookTorrance(n,l,v,h);

	return (diff+spec)*light.color*att*max(0.0,dot(n,l));

}

void main() {
	R0 = vec3(0.04);
	roughness = texture(roughnessTexture, vert.uv).r;
	metalness = texture(metalnessTexture, vert.uv).r;
	ao = texture(aoTexture, vert.uv).r;

	vec3 n = normalize(vert.normal_world);
	vec3 v = normalize(camera_world - vert.position_world);
	
	vec3 texColor = texture(diffuseTexture, vert.uv).rgb;
	R0 = mix(R0, texColor, metalness); //mix with metalness
	color = vec4(texColor * ao * vec3(0.03), 1); // ambient
	
	
	//Directional Light
	color.rgb += dirLight(dirL, n, v, texColor);

	//point Light
	color.rgb += ptLight(pointL, n, v, texColor);

	color.rgb = color.rgb / (color.rgb + vec3(1.0));
	color.rgb = pow(color.rgb, vec3(1.0/2.2));
}

